import {
  ApexAndNamedFilePaths,
  FileMetadata,
  FileMetadataGetter,
  FileMetadataSetter,
  FunctionInvocationFilePaths,
  LocalFileResolver
} from "upload-plugin-sdk";
import { promises as fsAsync } from "fs";
import path from "path";

export class TransformationLocalFileManager {
  private readonly apexFileName = "index";

  getMetadata(paths: FunctionInvocationFilePaths): FileMetadataGetter {
    return async (urlPath): Promise<FileMetadata> => {
      const metadataFilePath = this.resolveMetadataPath(paths, urlPath);
      if (!(await this.exists(metadataFilePath))) {
        return {};
      }
      return JSON.parse((await fsAsync.readFile(metadataFilePath)).toString());
    };
  }

  setMetadata(paths: FunctionInvocationFilePaths): FileMetadataSetter {
    return async (urlPath, metadata) => {
      const metadataFilePath = this.resolveMetadataPath(paths, urlPath);
      const dir = path.dirname(metadataFilePath);
      await fsAsync.mkdir(dir, { recursive: true });
      await fsAsync.writeFile(metadataFilePath, JSON.stringify(metadata));
    };
  }

  resolveFilePath(paths: FunctionInvocationFilePaths): LocalFileResolver {
    return urlPath => this.resolvePath(paths.fileData, urlPath);
  }

  private resolveMetadataPath(paths: FunctionInvocationFilePaths, urlPath: string): string {
    return this.resolvePath(paths.fileMetadata, urlPath);
  }

  /**
   * Q: Why use two DIRs instead of a single DIR and replace every "/" with a "/_" to support apex files?
   *
   * A: While the single directory approach is cleaner, it creates more work for plugins that operate against entire
   *    directories. For example, the upload-compression-plugin, which needs to extract files to a directory. If we used
   *    the single directory abstraction (which relies on renaming files to contain a "_" prefix), such plugins would need
   *    to somehow inject a path translation middleware into the code that determines where to (for example) extract files.
   *
   * Sync with: upload-transformer/TransformationFunctionInvoker/resolveFilePath
   */
  private resolvePath(root: ApexAndNamedFilePaths, urlPath: string): string {
    if (!urlPath.startsWith("/")) {
      throw new Error("URL path must start with '/'.");
    }

    return urlPath.endsWith("/")
      ? this.resolveFilePathSafely(root.apexFileDir, `${urlPath.substring(1)}${this.apexFileName}`)
      : this.resolveFilePathSafely(root.namedFileDir, urlPath.substring(1));
  }

  private resolveFilePathSafely(parentDir: string, childPath: string): string {
    const resolvedPath = path.resolve(parentDir, childPath);
    const isParentDirOrChild = resolvedPath === parentDir || resolvedPath.startsWith(parentDir);
    if (!isParentDirOrChild) {
      throw new Error(`URL path contained a file traversal (which is prohibited): '${childPath}'`);
    }
    return resolvedPath;
  }

  private async exists(folderOrFilePath: string): Promise<boolean> {
    try {
      await fsAsync.access(folderOrFilePath);
      return true;
    } catch {
      return false;
    }
  }
}
